@@ -6,10 +6,6 @@
  * See COPYING file for details.
  */
 
-/*
- * NOTES - make sure to run this as -s to avoid threading.
- * TODO - can we enforce that here from the code?
- */
 #define FUSE_USE_VERSION 26
 
 #include <stdio.h>
@@ -29,10 +25,13 @@
 #include <sys/mount.h>
 #include <wait.h>
 
+#include <nih-dbus/dbus_connection.h>
 #include <nih/alloc.h>
 #include <nih/string.h>
+#include <nih/error.h>
 
 #include "cgmanager.h"
+#include "config.h" // for VERSION
 
 struct lxcfs_state {
 	/*
@@ -43,6 +42,36 @@ struct lxcfs_state {
 };
 #define LXCFS_DATA ((struct lxcfs_state *) fuse_get_context()->private_data)
 
+enum {
+	LXC_TYPE_CGDIR,
+	LXC_TYPE_CGFILE,
+	LXC_TYPE_PROC_MEMINFO,
+	LXC_TYPE_PROC_CPUINFO,
+	LXC_TYPE_PROC_UPTIME,
+	LXC_TYPE_PROC_STAT,
+	LXC_TYPE_PROC_DISKSTATS,
+};
+
+struct file_info {
+	char *controller;
+	char *cgroup;
+	char *file;
+	int type;
+	char *buf;  // unused as of yet
+	int buflen;
+	int size; //actual data size
+};
+
+/* reserve buffer size, for cpuall in /proc/stat */
+#define BUF_RESERVE_SIZE 256
+
+static char *must_copy_string(void *parent, const char *str)
+{
+	if (!str)
+		return NULL;
+	return NIH_MUST( nih_strdup(parent, str) );
+}
+
 /*
  * TODO - return value should denote whether child exited with failure
  * so callers can return errors.  Esp read/write of tasks and cgroup.procs
@@ -313,6 +342,8 @@ static char *pick_controller_from_path(struct fuse_context *fc, const char *path
 
 	if (strlen(path) < 9)
 		return NULL;
+	if (*(path+7) != '/')
+		return NULL;
 	p1 = path+8;
 	ret = nih_strdup(NULL, p1);
 	if (!ret)
@@ -412,16 +443,6 @@ static void get_cgdir_and_path(const char *cg, char **dir, char **file)
 	*p = '\0';
 }
 
-static size_t get_file_size(const char *contrl, const char *cg, const char *f)
-{
-	nih_local char *data = NULL;
-	size_t s;
-	if (!cgm_get_value(contrl, cg, f, &data))
-		return -EINVAL;
-	s = strlen(data);
-	return s;
-}
-
 /*
  * FUSE ops for /cgroup
  */
@@ -514,7 +535,7 @@ static int cg_getattr(const char *path, struct stat *sb)
 		sb->st_nlink = 1;
 		sb->st_uid = k->uid;
 		sb->st_gid = k->gid;
-		sb->st_size = get_file_size(controller, path1, path2);
+		sb->st_size = 0;
 		return 0;
 	}
 
@@ -530,41 +551,59 @@ static int cg_opendir(const char *path, struct fuse_file_info *fi)
 	struct fuse_context *fc = fuse_get_context();
 	nih_local struct cgm_keys **list = NULL;
 	const char *cgroup;
+	struct file_info *dir_info;
 	nih_local char *controller = NULL;
-	nih_local char *nextcg = NULL;
 
 	if (!fc)
 		return -EIO;
 
-	if (strcmp(path, "/cgroup") == 0)
-		return 0;
+	if (strcmp(path, "/cgroup") == 0) {
+		cgroup = NULL;
+		controller = NULL;
+	} else {
+		// return list of keys for the controller, and list of child cgroups
+		controller = pick_controller_from_path(fc, path);
+		if (!controller)
+			return -EIO;
 
-	// return list of keys for the controller, and list of child cgroups
-	controller = pick_controller_from_path(fc, path);
-	if (!controller)
-		return -EIO;
-
-	cgroup = find_cgroup_in_path(path);
-	if (!cgroup) {
-		/* this is just /cgroup/controller, return its contents */
-		cgroup = "/";
+		cgroup = find_cgroup_in_path(path);
+		if (!cgroup) {
+			/* this is just /cgroup/controller, return its contents */
+			cgroup = "/";
+		}
 	}
 
-	if (!fc_may_access(fc, controller, cgroup, NULL, O_RDONLY))
+	if (cgroup && !fc_may_access(fc, controller, cgroup, NULL, O_RDONLY))
 		return -EACCES;
+
+	/* we'll free this at cg_releasedir */
+	dir_info = NIH_MUST( nih_alloc(NULL, sizeof(*dir_info)) );
+	dir_info->controller = must_copy_string(dir_info, controller);
+	dir_info->cgroup = must_copy_string(dir_info, cgroup);
+	dir_info->type = LXC_TYPE_CGDIR;
+	dir_info->buf = NULL;
+	dir_info->file = NULL;
+	dir_info->buflen = 0;
+
+	fi->fh = (unsigned long)dir_info;
 	return 0;
 }
 
 static int cg_readdir(const char *path, void *buf, fuse_fill_dir_t filler, off_t offset,
 		struct fuse_file_info *fi)
 {
+	struct file_info *d = (struct file_info *)fi->fh;
+	nih_local struct cgm_keys **list = NULL;
+	int i;
+	nih_local char *nextcg = NULL;
 	struct fuse_context *fc = fuse_get_context();
 
-	if (!fc)
+	if (d->type != LXC_TYPE_CGDIR) {
+		fprintf(stderr, "Internal error: file cache info used in readdir\n");
 		return -EIO;
-
-	if (strcmp(path, "/cgroup") == 0) {
-		// get list of controllers
+	}
+	if (!d->cgroup && !d->controller) {
+		// ls /var/lib/lxcfs/cgroup - just show list of controllers
 		char **list = LXCFS_DATA ? LXCFS_DATA->subsystems : NULL;
 		int i;
 
@@ -579,31 +618,11 @@ static int cg_readdir(const char *path, void *buf, fuse_fill_dir_t filler, off_t
 		return 0;
 	}
 
-	// return list of keys for the controller, and list of child cgroups
-	nih_local struct cgm_keys **list = NULL;
-	const char *cgroup;
-	nih_local char *controller = NULL;
-	int i;
-	nih_local char *nextcg = NULL;
-
-	controller = pick_controller_from_path(fc, path);
-	if (!controller)
-		return -EIO;
-
-	cgroup = find_cgroup_in_path(path);
-	if (!cgroup) {
-		/* this is just /cgroup/controller, return its contents */
-		cgroup = "/";
-	}
-
-	if (!fc_may_access(fc, controller, cgroup, NULL, O_RDONLY))
-		return -EACCES;
-
-	if (!cgm_list_keys(controller, cgroup, &list))
+	if (!cgm_list_keys(d->controller, d->cgroup, &list))
 		// not a valid cgroup
 		return -EINVAL;
 
-	if (!caller_is_in_ancestor(fc->pid, controller, cgroup, &nextcg)) {
+	if (!caller_is_in_ancestor(fc->pid, d->controller, d->cgroup, &nextcg)) {
 		if (nextcg) {
 			int ret;
 			ret = filler(buf, nextcg,  NULL, 0);
@@ -622,7 +641,7 @@ static int cg_readdir(const char *path, void *buf, fuse_fill_dir_t filler, off_t
 	// now get the list of child cgroups
 	nih_local char **clist = NULL;
 
-	if (!cgm_list_children(controller, cgroup, &clist))
+	if (!cgm_list_children(d->controller, d->cgroup, &clist))
 		return 0;
 	for (i = 0; clist[i]; i++) {
 		if (filler(buf, clist[i], NULL, 0) != 0) {
@@ -632,14 +651,23 @@ static int cg_readdir(const char *path, void *buf, fuse_fill_dir_t filler, off_t
 	return 0;
 }
 
+static void do_release_file_info(struct file_info *f)
+{
+	/*
+	 * all file_info fields which are nih_alloc()d with f as parent
+	 * will be automatically freed
+	 */
+	nih_free(f);
+}
+
 static int cg_releasedir(const char *path, struct fuse_file_info *fi)
 {
+	struct file_info *d = (struct file_info *)fi->fh;
+
+	do_release_file_info(d);
 	return 0;
 }
 
-/*
- * TODO - cache info here for read/write, release in cg_release.
- */
 static int cg_open(const char *path, struct fuse_file_info *fi)
 {
 	nih_local char *controller = NULL;
@@ -647,6 +675,7 @@ static int cg_open(const char *path, struct fuse_file_info *fi)
 	char *fpath = NULL, *path1, *path2;
 	nih_local char * cgdir = NULL;
 	nih_local struct cgm_keys *k = NULL;
+	struct file_info *file_info;
 	struct fuse_context *fc = fuse_get_context();
 
 	if (!fc)
@@ -668,15 +697,33 @@ static int cg_open(const char *path, struct fuse_file_info *fi)
 		path2 = fpath;
 	}
 
-	if ((k = get_cgroup_key(controller, path1, path2)) != NULL) {
-		if (!fc_may_access(fc, controller, path1, path2, fi->flags))
-			// should never get here
-			return -EACCES;
+	k = get_cgroup_key(controller, path1, path2);
+	if (!k)
+		return -EINVAL;
 
-		return 0;
-	}
+	if (!fc_may_access(fc, controller, path1, path2, fi->flags))
+		// should never get here
+		return -EACCES;
 
-	return -EINVAL;
+	/* we'll free this at cg_release */
+	file_info = NIH_MUST( nih_alloc(NULL, sizeof(*file_info)) );
+	file_info->controller = must_copy_string(file_info, controller);
+	file_info->cgroup = must_copy_string(file_info, path1);
+	file_info->file = must_copy_string(file_info, path2);
+	file_info->type = LXC_TYPE_CGFILE;
+	file_info->buf = NULL;
+	file_info->buflen = 0;
+
+	fi->fh = (unsigned long)file_info;
+	return 0;
+}
+
+static int cg_release(const char *path, struct fuse_file_info *fi)
+{
+	struct file_info *f = (struct file_info *)fi->fh;
+
+	do_release_file_info(f);
+	return 0;
 }
 
 static int msgrecv(int sockfd, void *buf, size_t len)
@@ -822,11 +869,11 @@ static void pid_to_ns(int sock, pid_t tpid)
 
 	while (recv_creds(sock, &cred, &v)) {
 		if (v == '1')
-			exit(0);
+			_exit(0);
 		if (write(sock, &cred.pid, sizeof(pid_t)) != sizeof(pid_t))
-			exit(1);
+			_exit(1);
 	}
-	exit(0);
+	_exit(0);
 }
 
 /*
@@ -844,21 +891,23 @@ static void pid_to_ns_wrapper(int sock, pid_t tpid)
 	fd_set s;
 	char v;
 
-	sprintf(fnam, "/proc/%d/ns/pid", tpid);
+	ret = snprintf(fnam, sizeof(fnam), "/proc/%d/ns/pid", tpid);
+	if (ret < 0 || ret >= sizeof(fnam))
+		_exit(1);
 	newnsfd = open(fnam, O_RDONLY);
 	if (newnsfd < 0)
-		exit(1);
+		_exit(1);
 	if (setns(newnsfd, 0) < 0)
-		exit(1);
+		_exit(1);
 	close(newnsfd);
 
 	if (pipe(cpipe) < 0)
-		exit(1);
+		_exit(1);
 
 loop:
 	cpid = fork();
 	if (cpid < 0)
-		exit(1);
+		_exit(1);
 
 	if (!cpid) {
 		char b = '1';
@@ -885,8 +934,8 @@ loop:
 	}
 
 	if (!wait_for_pid(cpid))
-		exit(1);
-	exit(0);
+		_exit(1);
+	_exit(0);
 
 again:
 	kill(cpid, SIGKILL);
@@ -992,52 +1041,41 @@ out:
 static int cg_read(const char *path, char *buf, size_t size, off_t offset,
 		struct fuse_file_info *fi)
 {
-	nih_local char *controller = NULL;
-	const char *cgroup;
-	char *fpath = NULL, *path1, *path2;
 	struct fuse_context *fc = fuse_get_context();
-	nih_local char * cgdir = NULL;
+	struct file_info *f = (struct file_info *)fi->fh;
 	nih_local struct cgm_keys *k = NULL;
 
-	if (offset)
+	if (f->type != LXC_TYPE_CGFILE) {
+		fprintf(stderr, "Internal error: directory cache info used in cg_read\n");
 		return -EIO;
+	}
+
+	if (offset)
+		return 0;
 
 	if (!fc)
 		return -EIO;
 
-	controller = pick_controller_from_path(fc, path);
-	if (!controller)
-		return -EINVAL;
-	cgroup = find_cgroup_in_path(path);
-	if (!cgroup)
+	if (!f->controller)
 		return -EINVAL;
 
-	get_cgdir_and_path(cgroup, &cgdir, &fpath);
-	if (!fpath) {
-		path1 = "/";
-		path2 = cgdir;
-	} else {
-		path1 = cgdir;
-		path2 = fpath;
-	}
-
-	if ((k = get_cgroup_key(controller, path1, path2)) != NULL) {
+	if ((k = get_cgroup_key(f->controller, f->cgroup, f->file)) != NULL) {
 		nih_local char *data = NULL;
 		int s;
 		bool r;
 
-		if (!fc_may_access(fc, controller, path1, path2, O_RDONLY))
+		if (!fc_may_access(fc, f->controller, f->cgroup, f->file, O_RDONLY))
 			// should never get here
 			return -EACCES;
 
-		if (strcmp(path2, "tasks") == 0 ||
-				strcmp(path2, "/tasks") == 0 ||
-				strcmp(path2, "/cgroup.procs") == 0 ||
-				strcmp(path2, "cgroup.procs") == 0)
+		if (strcmp(f->file, "tasks") == 0 ||
+				strcmp(f->file, "/tasks") == 0 ||
+				strcmp(f->file, "/cgroup.procs") == 0 ||
+				strcmp(f->file, "cgroup.procs") == 0)
 			// special case - we have to translate the pids
-			r = do_read_pids(fc->pid, controller, path1, path2, &data);
+			r = do_read_pids(fc->pid, f->controller, f->cgroup, f->file, &data);
 		else
-			r = cgm_get_value(controller, path1, path2, &data);
+			r = cgm_get_value(f->controller, f->cgroup, f->file, &data);
 
 		if (!r)
 			return -EINVAL;
@@ -1048,6 +1086,8 @@ static int cg_read(const char *path, char *buf, size_t size, off_t offset,
 		if (s > size)
 			s = size;
 		memcpy(buf, data, s);
+		if (s > 0 && s < size && data[s-1] != '\n')
+			buf[s++] = '\n';
 
 		return s;
 	}
@@ -1075,12 +1115,12 @@ static void pid_from_ns(int sock, pid_t tpid)
 		if (ret <= 0) {
 			fprintf(stderr, "%s: bad select before read from parent: %s\n",
 				__func__, strerror(errno));
-			exit(1);
+			_exit(1);
 		}
 		if ((ret = read(sock, &vpid, sizeof(pid_t))) != sizeof(pid_t)) {
 			fprintf(stderr, "%s: bad read from parent: %s\n",
 				__func__, strerror(errno));
-			exit(1);
+			_exit(1);
 		}
 		if (vpid == -1) // done
 			break;
@@ -1090,10 +1130,10 @@ static void pid_from_ns(int sock, pid_t tpid)
 			v = '1';
 			cred.pid = getpid();
 			if (send_creds(sock, &cred, v, false) != SEND_CREDS_OK)
-				exit(1);
+				_exit(1);
 		}
 	}
-	exit(0);
+	_exit(0);
 }
 
 static void pid_from_ns_wrapper(int sock, pid_t tpid)
@@ -1105,22 +1145,24 @@ static void pid_from_ns_wrapper(int sock, pid_t tpid)
 	struct timeval tv;
 	char v;
 
-	sprintf(fnam, "/proc/%d/ns/pid", tpid);
+	ret = snprintf(fnam, sizeof(fnam), "/proc/%d/ns/pid", tpid);
+	if (ret < 0 || ret >= sizeof(fnam))
+		_exit(1);
 	newnsfd = open(fnam, O_RDONLY);
 	if (newnsfd < 0)
-		exit(1);
+		_exit(1);
 	if (setns(newnsfd, 0) < 0)
-		exit(1);
+		_exit(1);
 	close(newnsfd);
 
 	if (pipe(cpipe) < 0)
-		exit(1);
+		_exit(1);
 
 loop:
 	cpid = fork();
 
 	if (cpid < 0)
-		exit(1);
+		_exit(1);
 
 	if (!cpid) {
 		char b = '1';
@@ -1148,8 +1190,8 @@ loop:
 	}
 
 	if (!wait_for_pid(cpid))
-		exit(1);
-	exit(0);
+		_exit(1);
+	_exit(0);
 
 again:
 	kill(cpid, SIGKILL);
@@ -1224,16 +1266,18 @@ out:
 int cg_write(const char *path, const char *buf, size_t size, off_t offset,
 	     struct fuse_file_info *fi)
 {
-	nih_local char *controller = NULL;
-	const char *cgroup;
-	char *fpath = NULL, *path1, *path2;
 	struct fuse_context *fc = fuse_get_context();
-	nih_local char * cgdir = NULL;
-	nih_local struct cgm_keys *k = NULL;
 	nih_local char *localbuf = NULL;
+	nih_local struct cgm_keys *k = NULL;
+	struct file_info *f = (struct file_info *)fi->fh;
+
+	if (f->type != LXC_TYPE_CGFILE) {
+		fprintf(stderr, "Internal error: directory cache info used in cg_write\n");
+		return -EIO;
+	}
 
 	if (offset)
-		return -EINVAL;
+		return 0;
 
 	if (!fc)
 		return -EIO;
@@ -1241,36 +1285,21 @@ int cg_write(const char *path, const char *buf, size_t size, off_t offset,
 	localbuf = NIH_MUST( nih_alloc(NULL, size+1) );
 	localbuf[size] = '\0';
 	memcpy(localbuf, buf, size);
-	controller = pick_controller_from_path(fc, path);
-	if (!controller)
-		return -EINVAL;
-	cgroup = find_cgroup_in_path(path);
-	if (!cgroup)
-		return -EINVAL;
 
-	get_cgdir_and_path(cgroup, &cgdir, &fpath);
-	if (!fpath) {
-		path1 = "/";
-		path2 = cgdir;
-	} else {
-		path1 = cgdir;
-		path2 = fpath;
-	}
-
-	if ((k = get_cgroup_key(controller, path1, path2)) != NULL) {
+	if ((k = get_cgroup_key(f->controller, f->cgroup, f->file)) != NULL) {
 		bool r;
 
-		if (!fc_may_access(fc, controller, path1, path2, O_WRONLY))
+		if (!fc_may_access(fc, f->controller, f->cgroup, f->file, O_WRONLY))
 			return -EACCES;
 
-		if (strcmp(path2, "tasks") == 0 ||
-				strcmp(path2, "/tasks") == 0 ||
-				strcmp(path2, "/cgroup.procs") == 0 ||
-				strcmp(path2, "cgroup.procs") == 0)
+		if (strcmp(f->file, "tasks") == 0 ||
+				strcmp(f->file, "/tasks") == 0 ||
+				strcmp(f->file, "/cgroup.procs") == 0 ||
+				strcmp(f->file, "cgroup.procs") == 0)
 			// special case - we have to translate the pids
-			r = do_write_pids(fc->pid, controller, path1, path2, localbuf);
+			r = do_write_pids(fc->pid, f->controller, f->cgroup, f->file, localbuf);
 		else
-			r = cgm_set_value(controller, path1, path2, localbuf);
+			r = cgm_set_value(f->controller, f->cgroup, f->file, localbuf);
 
 		if (!r)
 			return -EINVAL;
@@ -1495,24 +1524,24 @@ static void get_mem_cached(char *memstat, unsigned long *v)
 }
 
 static void get_blkio_io_value(char *str, unsigned major, unsigned minor, char *iotype, unsigned long *v)
-{   
+{
 	char *eol;
 	char key[32];
-	
+
 	memset(key, 0, 32);
 	snprintf(key, 32, "%u:%u %s", major, minor, iotype);
-	
+
 	size_t len = strlen(key);
 	*v = 0;
 
 	while (*str) {
 		if (startswith(str, key)) {
- 			sscanf(str + len, "%lu", v);
- 			return;
- 		}
- 		eol = strchr(str, '\n');
+			sscanf(str + len, "%lu", v);
+			return;
+		}
+		eol = strchr(str, '\n');
 		if (!eol)
- 			return;
+			return;
 		str = eol+1;
 	}
 }
@@ -1555,6 +1584,53 @@ out:
 	return answer;
 }
 
+static int read_file(const char *path, char *buf, size_t size,
+		     struct file_info *d)
+{
+	size_t linelen = 0, total_len = 0, rv = 0;
+	char *line = NULL;
+	char *cache = d->buf;
+	size_t cache_size = d->buflen;
+	FILE *f = fopen(path, "r");
+	if (!f)
+		return 0;
+
+	while (getline(&line, &linelen, f) != -1) {
+		size_t l = snprintf(cache, cache_size, "%s", line);
+		if (l < 0) {
+			perror("Error writing to cache");
+			rv = 0;
+			goto err;
+		}
+		if (l >= cache_size) {
+			fprintf(stderr, "Internal error: truncated write to cache\n");
+			rv = 0;
+			goto err;
+		}
+		if (l < cache_size) {
+			cache += l;
+			cache_size -= l;
+			total_len += l;
+		} else {
+			cache += cache_size;
+			total_len += cache_size;
+			cache_size = 0;
+			break;
+		}
+	}
+
+	d->size = total_len;
+	if (total_len > size ) total_len = size;
+
+	/* read from off 0 */
+	memcpy(buf, d->buf, total_len);
+	rv = total_len;
+  err:
+	fclose(f);
+	free(line);
+	return rv;
+}
+
 /*
  * FUSE ops for /proc
  */
@@ -1563,18 +1639,27 @@ static int proc_meminfo_read(char *buf, size_t size, off_t offset,
 		struct fuse_file_info *fi)
 {
 	struct fuse_context *fc = fuse_get_context();
+	struct file_info *d = (struct file_info *)fi->fh;
 	nih_local char *cg = get_pid_cgroup(fc->pid, "memory");
 	nih_local char *memlimit_str = NULL, *memusage_str = NULL, *memstat_str = NULL;
 	unsigned long memlimit = 0, memusage = 0, cached = 0, hosttotal = 0;
 	char *line = NULL;
-	size_t linelen = 0, total_len = 0;
+	size_t linelen = 0, total_len = 0, rv = 0;
+	char *cache = d->buf;
+	size_t cache_size = d->buflen;
 	FILE *f;
 
-	if (offset)
-		return -EINVAL;
+	if (offset){
+		if (offset > d->size)
+			return -EINVAL;
+		int left = d->size - offset;
+		total_len = left > size ? size: left;
+		memcpy(buf, cache + offset, total_len);
+		return total_len;
+	}
 
 	if (!cg)
-		return 0;
+		return read_file("/proc/meminfo", buf, size, d);
 
 	if (!cgm_get_value("memory", cg, "memory.limit_in_bytes", &memlimit_str))
 		return 0;
@@ -1620,15 +1705,34 @@ static int proc_meminfo_read(char *buf, size_t size, off_t offset,
 			printme = lbuf;
 		} else
 			printme = line;
-		l = snprintf(buf, size, "%s", printme);
-		buf += l;
-		size -= l;
+
+		l = snprintf(cache, cache_size, "%s", printme);
+		if (l < 0) {
+			perror("Error writing to cache");
+			rv = 0;
+			goto err;
+
+		}
+		if (l >= cache_size) {
+			fprintf(stderr, "Internal error: truncated write to cache\n");
+			rv = 0;
+			goto err;
+		}
+
+		cache += l;
+		cache_size -= l;
 		total_len += l;
 	}
 
+	d->size = total_len;
+	if (total_len > size ) total_len = size;
+	memcpy(buf, d->buf, total_len);
+
+	rv = total_len;
+  err:
 	fclose(f);
 	free(line);
-	return total_len;
+	return rv;
 }
 
 /*
@@ -1644,47 +1748,7 @@ static char *get_cpuset(const char *cg)
 	return answer;
 }
 
-/*
- * Helper functions for cpuset_in-set
- */
-char *cpuset_nexttok(const char *c)
-{
-	char *r = strchr(c+1, ',');
-	if (r)
-		return r+1;
-	return NULL;
-}
-
-int cpuset_getrange(const char *c, int *a, int *b)
-{
-	int ret;
-
-	ret = sscanf(c, "%d-%d", a, b);
-	return ret;
-}
-
-/*
- * cpusets are in format "1,2-3,4"
- * iow, comma-delimited ranges
- */
-static bool cpu_in_cpuset(int cpu, const char *cpuset)
-{
-	const char *c;
-
-	for (c = cpuset; c; c = cpuset_nexttok(c)) {
-		int a, b, ret;
-
-		ret = cpuset_getrange(c, &a, &b);
-		if (ret == 1 && cpu == a)
-			return true;
-		if (ret != 2) // bad cpuset!
-			return false;
-		if (cpu >= a && cpu <= b)
-			return true;
-	}
-
-	return false;
-}
+bool cpu_in_cpuset(int cpu, const char *cpuset);
 
 static bool cpuline_in_cpuset(const char *line, const char *cpuset)
 {
@@ -1711,19 +1775,28 @@ static int proc_cpuinfo_read(char *buf, size_t size, off_t offset,
 		struct fuse_file_info *fi)
 {
 	struct fuse_context *fc = fuse_get_context();
+	struct file_info *d = (struct file_info *)fi->fh;
 	nih_local char *cg = get_pid_cgroup(fc->pid, "cpuset");
 	nih_local char *cpuset = NULL;
 	char *line = NULL;
-	size_t linelen = 0, total_len = 0;
+	size_t linelen = 0, total_len = 0, rv = 0;
 	bool am_printing = false;
 	int curcpu = -1;
+	char *cache = d->buf;
+	size_t cache_size = d->buflen;
 	FILE *f;
 
-	if (offset)
-		return -EINVAL;
+	if (offset){
+		if (offset > d->size)
+			return -EINVAL;
+		int left = d->size - offset;
+		total_len = left > size ? size: left;
+		memcpy(buf, cache + offset, total_len);
+		return total_len;
+	}
 
 	if (!cg)
-		return 0;
+		return read_file("proc/cpuinfo", buf, size, d);
 
 	cpuset = get_cpuset(cg);
 	if (!cpuset)
@@ -1739,42 +1812,98 @@ static int proc_cpuinfo_read(char *buf, size_t size, off_t offset,
 			am_printing = cpuline_in_cpuset(line, cpuset);
 			if (am_printing) {
 				curcpu ++;
-				l = snprintf(buf, size, "processor	: %d\n", curcpu);
-				buf += l;
-				size -= l;
-				total_len += l;
+				l = snprintf(cache, cache_size, "processor	: %d\n", curcpu);
+				if (l < 0) {
+					perror("Error writing to cache");
+					rv = 0;
+					goto err;
+				}
+				if (l >= cache_size) {
+					fprintf(stderr, "Internal error: truncated write to cache\n");
+					rv = 0;
+					goto err;
+				}
+				if (l < cache_size){
+					cache += l;
+					cache_size -= l;
+					total_len += l;
+				}else{
+					cache += cache_size;
+					total_len += cache_size;
+					cache_size = 0;
+					break;
+				}
 			}
 			continue;
 		}
 		if (am_printing) {
-			l = snprintf(buf, size, "%s", line);
-			buf += l;
-			size -= l;
-			total_len += l;
+			l = snprintf(cache, cache_size, "%s", line);
+			if (l < 0) {
+				perror("Error writing to cache");
+				rv = 0;
+				goto err;
+			}
+			if (l >= cache_size) {
+				fprintf(stderr, "Internal error: truncated write to cache\n");
+				rv = 0;
+				goto err;
+			}
+			if (l < cache_size) {
+				cache += l;
+				cache_size -= l;
+				total_len += l;
+			} else {
+				cache += cache_size;
+				total_len += cache_size;
+				cache_size = 0;
+				break;
+			}
 		}
 	}
 
+	d->size = total_len;
+	if (total_len > size ) total_len = size;
+
+	/* read from off 0 */
+	memcpy(buf, d->buf, total_len);
+	rv = total_len;
+  err:
 	fclose(f);
 	free(line);
-	return total_len;
+	return rv;
 }
 
 static int proc_stat_read(char *buf, size_t size, off_t offset,
 		struct fuse_file_info *fi)
 {
 	struct fuse_context *fc = fuse_get_context();
+	struct file_info *d = (struct file_info *)fi->fh;
 	nih_local char *cg = get_pid_cgroup(fc->pid, "cpuset");
 	nih_local char *cpuset = NULL;
 	char *line = NULL;
-	size_t linelen = 0, total_len = 0;
+	size_t linelen = 0, total_len = 0, rv = 0;
 	int curcpu = -1; /* cpu numbering starts at 0 */
+	unsigned long user = 0, nice = 0, system = 0, idle = 0, iowait = 0, irq = 0, softirq = 0, steal = 0, guest = 0;
+	unsigned long user_sum = 0, nice_sum = 0, system_sum = 0, idle_sum = 0, iowait_sum = 0,
+					irq_sum = 0, softirq_sum = 0, steal_sum = 0, guest_sum = 0;
+#define CPUALL_MAX_SIZE BUF_RESERVE_SIZE
+	char cpuall[CPUALL_MAX_SIZE];
+	/* reserve for cpu all */
+	char *cache = d->buf + CPUALL_MAX_SIZE;
+	size_t cache_size = d->buflen - CPUALL_MAX_SIZE;
 	FILE *f;
 
-	if (offset)
-		return -EINVAL;
+	if (offset){
+		if (offset > d->size)
+			return -EINVAL;
+		int left = d->size - offset;
+		total_len = left > size ? size: left;
+		memcpy(buf, d->buf + offset, total_len);
+		return total_len;
+	}
 
 	if (!cg)
-		return 0;
+		return read_file("/proc/stat", buf, size, d);
 
 	cpuset = get_cpuset(cg);
 	if (!cpuset)
@@ -1784,6 +1913,12 @@ static int proc_stat_read(char *buf, size_t size, off_t offset,
 	if (!f)
 		return 0;
 
+	//skip first line
+	if (getline(&line, &linelen, f) < 0) {
+		fprintf(stderr, "proc_stat_read read first line failed\n");
+		goto out;
+	}
+
 	while (getline(&line, &linelen, f) != -1) {
 		size_t l;
 		int cpu;
@@ -1792,11 +1927,29 @@ static int proc_stat_read(char *buf, size_t size, off_t offset,
 
 		if (sscanf(line, "cpu%9[^ ]", cpu_char) != 1) {
 			/* not a ^cpuN line containing a number N, just print it */
-			l = snprintf(buf, size, "%s", line);
-			buf += l;
-			size -= l;
-			total_len += l;
-			continue;
+			l = snprintf(cache, cache_size, "%s", line);
+			if (l < 0) {
+				perror("Error writing to cache");
+				rv = 0;
+				goto err;
+			}
+			if (l >= cache_size) {
+				fprintf(stderr, "Internal error: truncated write to cache\n");
+				rv = 0;
+				goto err;
+			}
+			if (l < cache_size) {
+				cache += l;
+				cache_size -= l;
+				total_len += l;
+				continue;
+			} else {
+				//no more space, break it
+				cache += cache_size;
+				total_len += cache_size;
+				cache_size = 0;
+				break;
+			}
 		}
 
 		if (sscanf(cpu_char, "%d", &cpu) != 1)
@@ -1808,15 +1961,62 @@ static int proc_stat_read(char *buf, size_t size, off_t offset,
 		c = strchr(line, ' ');
 		if (!c)
 			continue;
-		l = snprintf(buf, size, "cpu%d %s", curcpu, c);
-		buf += l;
-		size -= l;
+		l = snprintf(cache, cache_size, "cpu%d%s", curcpu, c);
+		if (l < 0) {
+			perror("Error writing to cache");
+			rv = 0;
+			goto err;
+
+		}
+		if (l >= cache_size) {
+			fprintf(stderr, "Internal error: truncated write to cache\n");
+			rv = 0;
+			goto err;
+		}
+
+		cache += l;
+		cache_size -= l;
 		total_len += l;
+
+		if (sscanf(line, "%*s %lu %lu %lu %lu %lu %lu %lu %lu %lu", &user, &nice, &system, &idle, &iowait, &irq,
+			&softirq, &steal, &guest) != 9)
+			continue;
+		user_sum += user;
+		nice_sum += nice;
+		system_sum += system;
+		idle_sum += idle;
+		iowait_sum += iowait;
+		irq_sum += irq;
+		softirq_sum += softirq;
+		steal_sum += steal;
+		guest_sum += guest;
 	}
 
+	cache = d->buf;
+
+	int cpuall_len = snprintf(cpuall, CPUALL_MAX_SIZE, "%s %lu %lu %lu %lu %lu %lu %lu %lu %lu\n",
+		"cpu ", user_sum, nice_sum, system_sum, idle_sum, iowait_sum, irq_sum, softirq_sum, steal_sum, guest_sum);
+	if (cpuall_len > 0 && cpuall_len < CPUALL_MAX_SIZE){
+		memcpy(cache, cpuall, cpuall_len);
+		cache += cpuall_len;
+	}else{
+		/* shouldn't happen */
+		fprintf(stderr, "proc_stat_read copy cpuall failed, cpuall_len=%d\n", cpuall_len);
+		cpuall_len = 0;
+	}
+
+	memmove(cache, d->buf + CPUALL_MAX_SIZE, total_len);
+	total_len += cpuall_len;
+	d->size = total_len;
+	if (total_len > size ) total_len = size;
+
+	memcpy(buf, d->buf, total_len);
+  out:
+	rv = total_len;
+  err:
 	fclose(f);
 	free(line);
-	return total_len;
+	return rv;
 }
 
 /*
@@ -1855,7 +2055,10 @@ static long int get_pid1_time(pid_t pid)
 		return 0;
 	}
 
-	sprintf(fnam, "/proc/%d/ns/pid", pid);
+	ret = snprintf(fnam, sizeof(fnam), "/proc/%d/ns/pid", pid);
+	if (ret < 0 || ret >= sizeof(fnam))
+		return 0;
+
 	fd = open(fnam, O_RDONLY);
 	if (fd < 0) {
 		perror("get_pid1_time open of ns/pid");
@@ -1912,7 +2115,7 @@ loop:
 	}
 
 	wait_for_pid(cpid);
-	exit(0);
+	_exit(0);
 
 again:
 	kill(cpid, SIGKILL);
@@ -1937,7 +2140,7 @@ static long int getreaperage(pid_t qpid)
 		mtime = get_pid1_time(qpid);
 		if (write(mypipe[1], &mtime, sizeof(mtime)) != sizeof(mtime))
 			fprintf(stderr, "Warning: bad write from getreaperage\n");
-		exit(0);
+		_exit(0);
 	}
 
 	close(mypipe[1]);
@@ -1989,12 +2192,29 @@ static int proc_uptime_read(char *buf, size_t size, off_t offset,
 		struct fuse_file_info *fi)
 {
 	struct fuse_context *fc = fuse_get_context();
+	struct file_info *d = (struct file_info *)fi->fh;
 	long int reaperage = getreaperage(fc->pid);;
 	long int idletime = getprocidle();
+	size_t total_len = 0;
 
-	if (offset)
-		return -EINVAL;
-	return snprintf(buf, size, "%ld %ld\n", reaperage, idletime);
+	if (offset){
+		if (offset > d->size)
+			return -EINVAL;
+		return 0;
+	}
+
+	total_len = snprintf(buf, size, "%ld %ld\n", reaperage, idletime);
+	if (total_len < 0){
+		perror("Error writing to cache");
+		return 0;
+	}
+	if (total_len >= size){
+		d->size = size;
+		return size;
+	}
+
+	d->size = total_len;
+	return total_len;
 }
 
 static int proc_diskstats_read(char *buf, size_t size, off_t offset,
@@ -2002,6 +2222,7 @@ static int proc_diskstats_read(char *buf, size_t size, off_t offset,
 {
 	char dev_name[72];
 	struct fuse_context *fc = fuse_get_context();
+	struct file_info *d = (struct file_info *)fi->fh;
 	nih_local char *cg = get_pid_cgroup(fc->pid, "blkio");
 	nih_local char *io_serviced_str = NULL, *io_merged_str = NULL, *io_service_bytes_str = NULL,
 			*io_wait_time_str = NULL, *io_service_time_str = NULL;
@@ -2012,16 +2233,19 @@ static int proc_diskstats_read(char *buf, size_t size, off_t offset,
 	unsigned long ios_pgr = 0, tot_ticks = 0, rq_ticks = 0;
 	unsigned long rd_svctm = 0, wr_svctm = 0, rd_wait = 0, wr_wait = 0;
 	char *line = NULL;
-	size_t linelen = 0, total_len = 0;
+	size_t linelen = 0, total_len = 0, rv = 0;
 	unsigned int major = 0, minor = 0;
 	int i = 0;
 	FILE *f;
 
-	if (offset)
-		return -EINVAL;
+	if (offset){
+		if (offset > d->size)
+			return -EINVAL;
+		return 0;
+	}
 
 	if (!cg)
-		return 0;
+		return read_file("/proc/diskstats", buf, size, d);
 
 	if (!cgm_get_value("blkio", cg, "blkio.io_serviced", &io_serviced_str))
 		return 0;
@@ -2043,7 +2267,7 @@ static int proc_diskstats_read(char *buf, size_t size, off_t offset,
 		size_t l;
 		char *printme, lbuf[256];
 
-		i = sscanf(line, "%u %u %s", &major, &minor, dev_name);
+		i = sscanf(line, "%u %u %71s", &major, &minor, dev_name);
 		if(i == 3){
 			get_blkio_io_value(io_serviced_str, major, minor, "Read", &read);
 			get_blkio_io_value(io_serviced_str, major, minor, "Write", &write);
@@ -2053,7 +2277,7 @@ static int proc_diskstats_read(char *buf, size_t size, off_t offset,
 			read_sectors = read_sectors/512;
 			get_blkio_io_value(io_service_bytes_str, major, minor, "Write", &write_sectors);
 			write_sectors = write_sectors/512;
-			
+
 			get_blkio_io_value(io_service_time_str, major, minor, "Read", &rd_svctm);
 			rd_svctm = rd_svctm/1000000;
 			get_blkio_io_value(io_wait_time_str, major, minor, "Read", &rd_wait);
@@ -2074,7 +2298,7 @@ static int proc_diskstats_read(char *buf, size_t size, off_t offset,
 
 		memset(lbuf, 0, 256);
 		if (read || write || read_merged || write_merged || read_sectors || write_sectors || read_ticks || write_ticks) {
-			snprintf(lbuf, 256, "%u       %u %s %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu\n", 
+			snprintf(lbuf, 256, "%u       %u %s %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu\n",
 				major, minor, dev_name, read, read_merged, read_sectors, read_ticks,
 				write, write_merged, write_sectors, write_ticks, ios_pgr, tot_ticks, rq_ticks);
 			printme = lbuf;
@@ -2082,14 +2306,27 @@ static int proc_diskstats_read(char *buf, size_t size, off_t offset,
 			continue;
 
 		l = snprintf(buf, size, "%s", printme);
+		if (l < 0) {
+			perror("Error writing to fuse buf");
+			rv = 0;
+			goto err;
+		}
+		if (l >= size) {
+			fprintf(stderr, "Internal error: truncated write to cache\n");
+			rv = 0;
+			goto err;
+		}
 		buf += l;
 		size -= l;
 		total_len += l;
 	}
 
+	d->size = total_len;
+	rv = total_len;
+  err:
 	fclose(f);
 	free(line);
-	return total_len;
+	return rv;
 }
 
 static off_t get_procfile_size(const char *which)
@@ -2128,7 +2365,7 @@ static int proc_getattr(const char *path, struct stat *sb)
 			strcmp(path, "/proc/uptime") == 0 ||
 			strcmp(path, "/proc/stat") == 0 ||
 			strcmp(path, "/proc/diskstats") == 0) {
-		sb->st_size = get_procfile_size(path);
+		sb->st_size = 0;
 		sb->st_mode = S_IFREG | 00444;
 		sb->st_nlink = 1;
 		return 0;
@@ -2151,29 +2388,63 @@ static int proc_readdir(const char *path, void *buf, fuse_fill_dir_t filler, off
 
 static int proc_open(const char *path, struct fuse_file_info *fi)
 {
-	if (strcmp(path, "/proc/meminfo") == 0 ||
-			strcmp(path, "/proc/cpuinfo") == 0 ||
-			strcmp(path, "/proc/uptime") == 0 ||
-			strcmp(path, "/proc/stat") == 0 || 
-			strcmp(path, "/proc/diskstats") == 0)
-		return 0;
-	return -ENOENT;
+	int type = -1;
+	struct file_info *info;
+
+	if (strcmp(path, "/proc/meminfo") == 0)
+		type = LXC_TYPE_PROC_MEMINFO;
+	else if (strcmp(path, "/proc/cpuinfo") == 0)
+		type = LXC_TYPE_PROC_CPUINFO;
+	else if (strcmp(path, "/proc/uptime") == 0)
+		type = LXC_TYPE_PROC_UPTIME;
+	else if (strcmp(path, "/proc/stat") == 0)
+		type = LXC_TYPE_PROC_STAT;
+	else if (strcmp(path, "/proc/diskstats") == 0)
+		type = LXC_TYPE_PROC_DISKSTATS;
+	if (type == -1)
+		return -ENOENT;
+
+	info = NIH_MUST( nih_alloc(NULL, sizeof(*info)) );
+	memset(info, 0, sizeof(*info));
+	info->type = type;
+
+	info->buflen = get_procfile_size(path) + BUF_RESERVE_SIZE;
+	info->buf = NIH_MUST( nih_alloc(info, info->buflen) );
+	memset(info->buf, 0, info->buflen);
+	/* set actual size to buffer size */
+	info->size = info->buflen;
+
+	fi->fh = (unsigned long)info;
+	return 0;
+}
+
+static int proc_release(const char *path, struct fuse_file_info *fi)
+{
+	struct file_info *f = (struct file_info *)fi->fh;
+
+	do_release_file_info(f);
+	return 0;
 }
 
 static int proc_read(const char *path, char *buf, size_t size, off_t offset,
 		struct fuse_file_info *fi)
 {
-	if (strcmp(path, "/proc/meminfo") == 0)
+	struct file_info *f = (struct file_info *) fi->fh;
+
+	switch (f->type) {
+	case LXC_TYPE_PROC_MEMINFO:
 		return proc_meminfo_read(buf, size, offset, fi);
-	if (strcmp(path, "/proc/cpuinfo") == 0)
+	case LXC_TYPE_PROC_CPUINFO:
 		return proc_cpuinfo_read(buf, size, offset, fi);
-	if (strcmp(path, "/proc/uptime") == 0)
+	case LXC_TYPE_PROC_UPTIME:
 		return proc_uptime_read(buf, size, offset, fi);
-	if (strcmp(path, "/proc/stat") == 0)
+	case LXC_TYPE_PROC_STAT:
 		return proc_stat_read(buf, size, offset, fi);
-	if (strcmp(path, "/proc/diskstats") == 0)
+	case LXC_TYPE_PROC_DISKSTATS:
 		return proc_diskstats_read(buf, size, offset, fi);
-	return -EINVAL;
+	default:
+		return -EINVAL;
+	}
 }
 
 /*
@@ -2277,7 +2548,12 @@ static int lxcfs_flush(const char *path, struct fuse_file_info *fi)
 
 static int lxcfs_release(const char *path, struct fuse_file_info *fi)
 {
-	return 0;
+	if (strncmp(path, "/cgroup", 7) == 0)
+		return cg_release(path, fi);
+	if (strncmp(path, "/proc", 5) == 0)
+		return proc_release(path, fi);
+
+	return -EINVAL;
 }
 
 static int lxcfs_fsync(const char *path, int datasync, struct fuse_file_info *fi)
@@ -2374,7 +2650,8 @@ static void usage(const char *me)
 {
 	fprintf(stderr, "Usage:\n");
 	fprintf(stderr, "\n");
-	fprintf(stderr, "%s [FUSE and mount options] mountpoint\n", me);
+	fprintf(stderr, "%s mountpoint\n", me);
+	fprintf(stderr, "%s -h\n", me);
 	exit(1);
 }
 
@@ -2388,25 +2665,103 @@ static bool is_help(char *w)
 	return false;
 }
 
+void swallow_arg(int *argcp, char *argv[], char *which)
+{
+	int i;
+
+	for (i = 1; argv[i]; i++) {
+		if (strcmp(argv[i], which) != 0)
+			continue;
+		for (; argv[i]; i++) {
+			argv[i] = argv[i+1];
+		}
+		(*argcp)--;
+		return;
+	}
+}
+
+void swallow_option(int *argcp, char *argv[], char *opt, char *v)
+{
+	int i;
+
+	for (i = 1; argv[i]; i++) {
+		if (!argv[i+1])
+			continue;
+		if (strcmp(argv[i], opt) != 0)
+			continue;
+		if (strcmp(argv[i+1], v) != 0) {
+			fprintf(stderr, "Warning: unexpected fuse option %s\n", v);
+			exit(1);
+		}
+		for (; argv[i+1]; i++) {
+			argv[i] = argv[i+2];
+		}
+		(*argcp) -= 2;
+		return;
+	}
+}
+
+bool detect_libnih_threadsafe(void)
+{
+#ifdef HAVE_NIH_THREADSAFE
+	if (nih_threadsafe())
+		return true;
+#endif
+	return false;
+}
+
 int main(int argc, char *argv[])
 {
-	int ret;
-	struct lxcfs_state *d;
+	int ret = -1;
+	struct lxcfs_state *d = NULL;
+	/*
+	 * what we pass to fuse_main is:
+	 * argv[0] -s -f -o allow_other,directio argv[1] NULL
+	 */
+	int nargs = 6;
+	bool threadsafe = detect_libnih_threadsafe();
+	char *newargv[7]; // one more than if needed if threadsafe
 
-	if (argc < 2 || is_help(argv[1]))
+	threadsafe = false; // still not safe with libnih+libdbus
+
+	dbus_threads_init_default();
+
+	if (threadsafe)
+		nargs = 5;
+
+	/* accomodate older init scripts */
+	swallow_arg(&argc, argv, "-s");
+	swallow_arg(&argc, argv, "-f");
+	swallow_option(&argc, argv, "-o", "allow_other");
+
+	if (argc == 2  && strcmp(argv[1], "--version") == 0) {
+		fprintf(stderr, "%s\n", VERSION);
+		exit(0);
+	}
+	if (argc != 2 || is_help(argv[1]))
 		usage(argv[0]);
 
-	d = malloc(sizeof(*d));
-	if (!d)
-		return -1;
+	d = NIH_MUST( malloc(sizeof(*d)) );
+
+	int cnt = 0;
+	newargv[cnt++] = argv[0];
+	if (!threadsafe)
+		newargv[cnt++] = "-s";
+	newargv[cnt++] = "-f";
+	newargv[cnt++] = "-o";
+	newargv[cnt++] = "allow_other,direct_io";
+	newargv[cnt++] = argv[1];
+	newargv[cnt++] = NULL;
 
 	if (!cgm_escape_cgroup())
 		fprintf(stderr, "WARNING: failed to escape to root cgroup\n");
 
 	if (!cgm_get_controllers(&d->subsystems))
-		return -1;
+		goto out;
 
-	ret = fuse_main(argc, argv, &lxcfs_ops, d);
+	ret = fuse_main(nargs, newargv, &lxcfs_ops, d);
 
+out:
+	free(d);
 	return ret;
 }
